/*
  Copyright (C) 2001-2002 Laurent Martelli.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

package org.objectweb.jac.aspects.persistence;


import java.util.*;
import org.objectweb.jac.core.*;
import org.objectweb.jac.util.*;

/**
 * Base class for collection wrappers
 */

public abstract class CollectionWrapper extends PersistenceWrapper {
   boolean isLoaded = false;

   public CollectionWrapper(AspectComponent ac, boolean isLoaded) {
      super(ac);
      this.isLoaded = isLoaded;
   }

   /**
    * Load the whole collection if it is not already loaded
    */
   public synchronized void load(Wrappee wrappee) throws Exception {
      if (!isLoaded) {
         Log.trace("persistence","loading collection "+getOID(wrappee)+" - "+wrappee);
         doLoad(wrappee);
         isLoaded = true;
      }
   }

   public boolean isLoaded() {
      return isLoaded;
   }

   /**
    * Unload the collection.
    */
   public synchronized void unload(Wrappee wrappee) {
      Log.trace("persistence",getOID(wrappee)+".unload...");
      isLoaded=false;
      Wrapping.invokeOrg(wrappee,"clear",new Object[0]);      
   }

   /**
    * Really load the whole collection. This is an abstract method
    * must be overriden by subclasses.
    */
   protected abstract void doLoad(Wrappee wrappee) throws Exception;

   /**
    * Remove all instances from the collection
    */
   public abstract Object clear(Interaction interaction) throws Exception;

   public Object preload(Interaction interaction) throws Exception {
      Log.trace("persistence",getOID(interaction.wrappee)+
                ".preload for "+interaction.method);
      try {
         load(interaction.wrappee);
      } catch (Exception e) {
         Log.warning("persistence","Failed to collection for "+interaction+": "+e);
         e.printStackTrace();
      }
      return proceed(interaction);
   }
   
   public Object size(Interaction interaction) throws Exception {
      if (!isLoaded) {
         return new Integer(
            new Long(getCollectionSize(getOID(interaction.wrappee))).intValue() );
      } else {
         return proceed(interaction);
      }
   }

   protected abstract long getCollectionSize(OID oid) throws Exception;

   public Object isEmpty(Interaction interaction) throws Exception {
      if (!isLoaded) {
         return ExtBoolean.valueOf(getCollectionSize(getOID(interaction.wrappee))==0);
      } else {
         return proceed(interaction);
      }
   }

   Date useDate=new Date();

   public Object memorizeUseDate(Interaction i) {
      useDate=new Date();
      return proceed(i);
   }

   public Date getUseDate(Wrappee wrappee) {
      return useDate;
   }

   public abstract Object iterator(Interaction interaction);
}

