/*
  Copyright (C) 2002 Laurent Martelli <laurent@aopsys.com>, 
  Julien van Malderen <julien@aopsys.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA */

package org.objectweb.jac.aspects.persistence;

import org.objectweb.jac.util.Log;
import org.objectweb.jac.util.Base64;
import org.objectweb.jac.core.rtti.ClassRepository;

/**
 * General converter for all Objects.
 *
 * <p>It converts Objects into Strings to store them for persistence,
 * and converts stored Strings into their original form for new
 * use.</p>
 */

public class ValueConverter
{

   /**
    * Returns a string representation of a field so that it can be
    * stored.<p>
    *
    * @param obj a persistent or primitive object
    * @return a ready to store string representation
    */
   static public String objectToString(Object obj)
   {
      if( obj == null ) return "null";
      Class lClass=obj.getClass();
      String className=lClass.getName();
      String value;
      
      StringConverter converter = (StringConverter)ClassRepository.get()
         .getClass(className).getAttribute(PersistenceAC.VALUE_CONVERTER);

      if (obj instanceof OID)
         value = ((OID)obj).toString();
      else {
         if (lClass == java.lang.String.class) {
            value = (String)obj; 
         } else if (converter != null) {
            value = converter.objectToString(obj);
         } else if (lClass.isArray() &&
                    (lClass.getComponentType()==byte.class)) {
            // array of bytes
            value = Base64.encodeToString((byte[])obj);
         }
         else {
            value = obj.toString();
         }
         value = className+":"+value;
      }
      return value;
   }

   /**
    * Returns an object from a string, depending on the needed type.<p>
    *
    * @param str the type and value in a string format (type:value)
    * @return an object value deduced from the string representation
    * and from the needed type
    */
   static public Object stringToObject(String str)
   {
      Log.trace("ValueConverter","stringToObject("+str+")");
      if ( str.equals("null") ) return null;
      char firstChar = str.charAt(0);
      if (Character.isDigit(firstChar)) {
         return new LongOID(Long.parseLong(str));
      } else {
         int separator = str.indexOf(":");
         String type = str.substring(0,separator);
         str = str.substring(separator+1);
         Log.trace("storage.conv","type = "+type+" ; value = "+str);

         StringConverter converter = (StringConverter)ClassRepository.get()
            .getClass(type).getAttribute(PersistenceAC.VALUE_CONVERTER);

         if (type.equals("java.lang.String")) {
            return str;
         } else if (converter != null) {
            return converter.stringToObject(str);
         } else if (type.equals("boolean") ||
                    type.equals("java.lang.Boolean")) {
            if (str.equals("true"))
               return Boolean.TRUE;
            else if (str.equals("false"))
               return Boolean.FALSE;
            else {
               Log.error("BAD BOOLEAN VALUE "+str);
               return null;
            }
         } else if (type.equals("java.io.File")) {
            return new java.io.File(str);
         } else if (type.equals("int") || type.equals("java.lang.Integer")) {
            return new Integer(str);
         } else if (type.equals("long") || type.equals("java.lang.Long")) {
            return new Long(str);
         } else if (type.equals("float") || type.equals("java.lang.Float")) {
            return new Float(str);
         } else if (type.equals("double") || type.equals("java.lang.Double")) {
            return new Double(str);
         } else if (type.equals("byte") || type.equals("java.lang.Byte")) {
            return new Byte(str);
         } else if (type.equals("[B")) {
            return Base64.decode(str);
         }
         else {
            Log.error("Unhandled type for value `"+str+"' : "+type);
            throw new Error("Unhandled type for value `"+str+"' : "+type);
         }
      }
   }

}
